﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace CSharpDemo
{
    class Duxbury
    {
        long[] engineState = new long[2];
        IntPtr engineHandle;
        string lastFailure = "";
        private static string tablePath;   // Path to table files
        private static string tablePrefix; // Like tablePath, but always ends in path separator
        private char[] lastTranslation = null;
        private UInt16[] lastInputPositions = null;

        public Duxbury(string brailleTable, string scanTable, bool unicode, bool in28xx, bool out28xx)
        {
            engineHandle = CreateEngine(brailleTable, scanTable, unicode, in28xx, out28xx, ref engineState);
            string assemblyPath = System.Reflection.Assembly.GetCallingAssembly().Location;
        }

        ~Duxbury()
        {
            Destroy(engineHandle);
        }

        public static string TablePath
        {
            get
            {
                return tablePath;
            }
            set
            {
                tablePath = value;
                if (value.EndsWith("\\"))
                {
                    tablePrefix = tablePath;
                }
                else
                {
                    tablePrefix = tablePath + "\\";
                }

            }
        }

        public bool IsOk
        {
            get
            {
                return engineHandle != IntPtr.Zero;
            }
        }

        public string LastFailure
        {
            get
            {
                return lastFailure;
            }
        }

        unsafe private IntPtr CreateEngine(string brailleTable, string scanTable, bool unicode, bool in28xx, bool out28xx)
        {
            return CreateEngine(brailleTable, scanTable, unicode, in28xx, out28xx, null);
        }

        unsafe private IntPtr CreateEngine(string brailleTable, string scanTable, bool unicode, bool in28xx, bool out28xx, ref long[] state)
        {
            fixed (long* statePtr = &state[0])
            {
                return CreateEngine(brailleTable, scanTable, unicode, in28xx, out28xx, statePtr);
            }
        }

        private bool CheckFunctionResult(UInt16 functionResult, string function)
        {
            if (functionResult != 0)
            {
                lastFailure = function + "() -> " + ErrorMessage(functionResult);
            }
            // TODO: Log, in case of failure
            return (functionResult == 0);
        }

        public string ErrorMessage(UInt16 functionResult)
        {
            switch (functionResult)
            {
                case 0: return "BRLTRN_OK";
                case 1: return "BRLTRN_ERROR_CANNOT_OPEN_BTB";       /* Failure when opening .BTB file */
                case 2: return "BRLTRN_ERROR_CANNOT_INITIALIZE_BTE"; /* Failure when initializing translator */
                case 3: return "BRLTRN_ERROR_CANNOT_OPEN_SCT";       /* Failure when opening .SCT file */
                case 4: return "BRLTRN_ERROR_CANNOT_INITIALIZE_NIN"; /* Failure when initializing scanner */
                case 5: return "BRLTRN_ERROR_CANNOT_READ";           /* Cannot init. input buffer stream */
                case 6: return "BRLTRN_ERROR_CANNOT_RESET_NIN";      /* Cannot reset scanner for new input */
                case 7: return "BRLTRN_ERROR_CANNOT_WRITE";          /* Cannot init. ouptut buffer stream */
                case 8: return "BRLTRN_ERROR_TRANSLATION_FAILED";    /* Generic error while actually translating */
                case 9: return "BRLTRN_ERROR_LICENSE_NOT_FOUND";     /* Cannot open license file */
                case 10: return "BRLTRN_ERROR_LICENSE_INVALID";       /* License file not valid */
                case 11: return "BRLTRN_ERROR_CANNOT_OPEN_CHITAB";    /* Cannot open chitab.txt */
                case 12: return "BRLTRN_ERROR_CANNOT_ALLOCATE_MAP";   /* Cannot allocate space for WinANSI -> dusci map */
                case 13: return "BRLTRN_ERROR_CANNOT_RESET_BTE";      /* Cannot reset Braille translation engine */
                case 14: return "BRLTRN_ERROR_INVALID_PARAMETER";     /* Invalid or NULL parameter */
                case 15: return "BRLTRN_ERROR_CANNOT_OPEN_SCRUB";     /* Cannot open SCRUB table (.SCT file) */
                case 16: return "BRLTRN_ERROR_TRANSLATION_NOT_READY"; /* BRLTRN_CopyDynamicResults called before results generated */

                case 22: return "BRLTRN_ERROR_BAD_VARIANT_TYPE"; /* All variant types passed by ref should be of type BSTR, else this error */
                case 23: return "BRLTRN_ERROR_OUT_OF_MEMORY";
                default: return functionResult.ToString();
            }
        }

        private static void Destroy(IntPtr engineHandle)
        {
            if (engineHandle != IntPtr.Zero)
            {
                if (IntPtr.Size == 4)
                    Duxbury32.BRLTRN_Destroy(engineHandle);
                else
                    Duxbury64.BRLTRN_Destroy(engineHandle);
            }
        }

        unsafe private IntPtr CreateEngine(string brailleTable, string scanTable, bool unicode, bool in28xx, bool out28xx, Int64* state)
        {
            string appPath = System.IO.Directory.GetCurrentDirectory();
            string licenseFilePath = null;
            string chitabFilePath = tablePrefix + "chitab.txt"; // System.IO.Path.Combine(appPath, @"duxbury\chitab.txt");
            UInt16 functionResult = 0;
            IntPtr engineHandle = IntPtr.Zero;

            if (unicode)
            {
                if (IntPtr.Size == 4)
                    functionResult = Duxbury32.BRLTRN_CreateW(tablePrefix + brailleTable, tablePrefix + scanTable, ref engineHandle, licenseFilePath, chitabFilePath);
                else
                    functionResult = Duxbury64.BRLTRN_CreateW(tablePrefix + brailleTable, tablePrefix + scanTable, ref engineHandle, licenseFilePath, chitabFilePath);
                if (!CheckFunctionResult(functionResult, "BRLTRN_CreateW")) return IntPtr.Zero;
            }
            else
            {
                if (IntPtr.Size == 4)
                    functionResult = Duxbury32.BRLTRN_CreateA(tablePrefix + brailleTable, tablePrefix + scanTable, ref engineHandle, licenseFilePath, chitabFilePath);
                else
                    functionResult = Duxbury64.BRLTRN_CreateA(tablePrefix + brailleTable, tablePrefix + scanTable, ref engineHandle, licenseFilePath, chitabFilePath);
                if (!CheckFunctionResult(functionResult, "BRLTRN_CreateA")) return IntPtr.Zero;
            }

            if (state != null)
            {
                if (IntPtr.Size == 4)
                    functionResult = Duxbury32.BRLTRN_SetStateBitmapVariableArray(engineHandle, state);
                else
                    functionResult = Duxbury64.BRLTRN_SetStateBitmapVariableArray(engineHandle, state);
                if (!CheckFunctionResult(functionResult, "BRLTRN_SetStateBitmapVariableArray")) { Destroy(engineHandle); return IntPtr.Zero; }
            }

            if (IntPtr.Size == 4)
                functionResult = Duxbury32.BRLTRN_Map2800(engineHandle, in28xx ? 1 : 0, out28xx ? 1 : 0);
            else
                functionResult = Duxbury64.BRLTRN_Map2800(engineHandle, in28xx ? 1 : 0, out28xx ? 1 : 0);
            if (!CheckFunctionResult(functionResult, "BRLTRN_Map2800")) { Destroy(engineHandle); return IntPtr.Zero; }

            UInt16 stateArraySize = 0;
            if (IntPtr.Size == 4)
                functionResult = Duxbury32.BRLTRN_GetStateArraySize(out stateArraySize);
            else
                functionResult = Duxbury64.BRLTRN_GetStateArraySize(out stateArraySize);
            if (!CheckFunctionResult(functionResult, "BRLTRN_GetStateArraySize")) { Destroy(engineHandle); return IntPtr.Zero; }

            return engineHandle;
        }

        public string TranslateStringW(string input, UInt16 inputCursorPos, ref UInt16 outputCursorPos)
        {
            UInt16 functionResult = 0;
            char[] outbuff = new char[input.Length * 4];
            UInt16 outputSize = (UInt16)(input.Length * 4);

            if (IntPtr.Size == 4)
                functionResult = Duxbury32.BRLTRN_TranslateStringW(engineHandle, input, outbuff, outputSize, inputCursorPos, ref outputCursorPos);
            else
                functionResult = Duxbury64.BRLTRN_TranslateStringW(engineHandle, input, outbuff, outputSize, inputCursorPos, ref outputCursorPos);

            if (!CheckFunctionResult(functionResult, "BRLTRN_TranslateString")) return "";
            return CharBufferToString(outbuff, 0, input.Length * 4);
        }

	private string CharBufferToString(char[] buffer, int length, int size)
	{
	    if (length == 0)
	    {
            for (int i = 0; i < size; i++)
                if (buffer[i] == '\0')
                {
                    length = i;
                    break;
                }
		}
            return new string(buffer.Take(length).ToArray());
	}

        public string TranslateStringA(string input, UInt16 inputCursorPos, ref UInt16 outputCursorPos)
        {
            UInt16 functionResult = 0;
            byte[] outbuff = new byte[input.Length * 4];
            char[] outbuffChar = new char[input.Length * 4];
            UInt16 outputSize = (UInt16)(input.Length * 4);

            if (IntPtr.Size == 4)
                functionResult = Duxbury32.BRLTRN_TranslateStringA(engineHandle, input, outbuff, outputSize, inputCursorPos, ref outputCursorPos);
            else
                functionResult = Duxbury64.BRLTRN_TranslateStringA(engineHandle, input, outbuff, outputSize, inputCursorPos, ref outputCursorPos);

            if (!CheckFunctionResult(functionResult, "BRLTRN_TranslateString")) return "";

            int outputLength = 0;
            for (int i = 0; i < outputSize; i++)
            {
                outbuffChar[i] = (char) outbuff[i];
                if (outbuff[i] == '\0')
                {
                    outputLength = i;
                    break;
                }
            }
            return new string(outbuffChar.Take(outputLength).ToArray());
        }

        public string TranslateStringW(string input)
        {
            UInt16 outputCursorPos = 0; // Dummy variable
            return TranslateStringW(input, 0, ref outputCursorPos);
        }

        public string TranslateStringA(string input)
        {
            UInt16 outputCursorPos = 0; // Dummy variable
            return TranslateStringA(input, 0, ref outputCursorPos);
        }

        public string TranslateStringEx(string input, int fixedOutputSize)
        {
            char[] translation = null;
            UInt16[] inputPositions = null;

               if (fixedOutputSize > 0)
            {
                translation = new char[fixedOutputSize];
                inputPositions = new UInt16[fixedOutputSize];
            }

            UInt16 result;
            UInt16 outputLen = 0;
            if (IntPtr.Size == 4)
            {
                result = Duxbury32.BRLTRN_TranslateStringEx(engineHandle, null, input, translation, inputPositions, (UInt16) fixedOutputSize, ref outputLen);
            }
            else
            {
                result = Duxbury64.BRLTRN_TranslateStringEx(engineHandle, null, input, translation, inputPositions, (UInt16) fixedOutputSize, ref outputLen);
            }
            if (!CheckFunctionResult(result, "BRLTRN_TranslateEx")) return "";

                if (fixedOutputSize > 0)
                {
                    lastTranslation = translation;
                    lastInputPositions = inputPositions;
                }
                else
                {
                UInt16 outputSize = (UInt16) (outputLen + 1);
                lastTranslation = new char[outputSize];
                lastInputPositions = new UInt16[outputSize];
                if (IntPtr.Size == 4)
                {
                    result = Duxbury32.BRLTRN_CopyDynamicResults(engineHandle, lastTranslation, lastInputPositions, outputSize);
                }
                else
                {
                       result = Duxbury64.BRLTRN_CopyDynamicResults(engineHandle, lastTranslation, lastInputPositions, outputSize);
                }
                if (!CheckFunctionResult(result, "BRLTRN_FetchDynamicResults"))
                    {
                        lastTranslation = null;
                        lastInputPositions = null;
                        return "";
                    }
                }
                return CharBufferToString(lastTranslation, outputLen, outputLen + 1);
        }

        public string TranslateStringEx(string input)
        {
            return TranslateStringEx(input, 0);
        }

        public Int16 LocateInputPos(Int16 outputPos)
        {
            if (lastInputPositions == null)
            {
                return 0;
            }
            if (outputPos < 0 || outputPos > lastInputPositions.Length)
            {
                return 0;
            }
            return (short) lastInputPositions[outputPos];
        }

        public Int16 InputPosFromOutputPos(string input, Int16 outputPos)
        {
            ushort functionResult;
            ushort inputPos = 0;

            if (IntPtr.Size == 4)
                functionResult = Duxbury32.BRLTRN_InputPosFromOutputPos(engineHandle, input, (ushort)outputPos, ref inputPos);
            else
                functionResult = Duxbury64.BRLTRN_InputPosFromOutputPos(engineHandle, input, (ushort)outputPos, ref inputPos);

            if (!CheckFunctionResult(functionResult, "BRLTRN_InputPosFromOutputPos")) return -1;

            if (inputPos < 0 && outputPos <= 1) inputPos = 0;
            return (Int16)inputPos;
        }

    }
}
